module sg_model_mod
	use sg_model_type_mod
	use sg_distrib_type_mod
	use sg_io_mod
	use sep
	use sep3d_struct_mod
	use sep_timers_mod
	use adj_mod

	implicit none

	integer							:: tm_u4, tm_v4, tm_tau4, tm_sp, tm_space, tm_write, tm_read
	integer, external, private		:: omp_get_thread_num, omp_get_num_threads
	integer, private, save			:: nrec, nsrc
	integer, private, save			:: one_step_nthreads
	real							:: eps
	type (filter)					:: dfilt
	integer							:: ncoeffs
	type (model_type)				:: m_s
	type (distrib_type)				:: d_s
	type (io_type)					:: io_s
	real							:: idt, idx, idz, dt_glob

contains

subroutine model_init (m_s_in, tag)
	type (model_type)	:: m_s_in
	integer, external	:: sreed
	integer				:: n, irec, isrc
	real				:: d
	type (sep3d)		:: mod_s
	character (len=*)	:: tag


	call from_param ("adj", m_s_in%adj, .false.)
	call from_param ("add", m_s_in%add, .false.)

	! read dimensions from one of the medium model files
	call init_sep3d (tag, mod_s, "INPUT")
	m_s_in%nx = mod_s%n(1)
	m_s_in%nz = mod_s%n(2)
	m_s_in%dx = mod_s%d(1)
	m_s_in%dz = mod_s%d(2)

	call mk_sponge(m_s_in)

	!expand to include sponge size
	m_s_in%nz = m_s_in%nz + m_s_in%nsponge_z_l + m_s_in%nsponge_z_u
	m_s_in%nx = m_s_in%nx + m_s_in%nsponge_x_l + m_s_in%nsponge_x_u

	! first index. This is a leftover from the MPI version
	m_s_in%fx = 1
	m_s_in%fz = 1

	call from_param("nt",m_s_in%nt)
	call from_param("dt",m_s_in%dt)
	call from_param("fpeak",m_s_in%fpeak,50.)
	call make_wavelet(m_s_in)

	! point source type. 0=pressure. 1=vertical. 3=horizontal.
	call from_param("isrc_type", m_s_in%isrc_type, 0)
	call from_param("pstrength", m_s_in%Pstrength, 1e+5)

	! point source location. default: at center of unpadded field
	call from_param("ixsrc", m_s_in%ixsrc, m_s_in%nx/2 - m_s_in%nsponge_x_l - m_s_in%nsponge_x_u)
	call from_param("izsrc", m_s_in%izsrc, m_s_in%nz/2 - m_s_in%nsponge_z_l - m_s_in%nsponge_z_u)

	! shift source location to padded field coordinates
	m_s_in%ixsrc = m_s_in%ixsrc + m_s_in%nsponge_x_l
	m_s_in%izsrc = m_s_in%izsrc + m_s_in%nsponge_z_l

	! input wavelet from point source or from virtual source in the forward modeling operation
	call from_param ("proptype", m_s_in%proptype, source)

	call from_param ("one_step_nthreads", one_step_nthreads, 1)
	call omp_set_num_threads (one_step_nthreads)

	!##### read receiver locations
	call from_aux("recs","n1",n)
	if(n/=2) call seperr("expecting receiver file to have n1 = 2")

	call from_aux("recs","n2",nrec)
	allocate(m_s_in%recs(2,nrec))

	if (4 * size (m_s_in%recs) /= sreed ("recs", m_s_in%recs, 4 * size (m_s_in%recs))) &
		call seperr("model_mod: trouble reading in recs")

	do irec = 1, nrec
		m_s_in%recs(1,irec)=m_s_in%recs(1,irec)+m_s_in%nsponge_x_l
		m_s_in%recs(2,irec)=m_s_in%recs(2,irec)+m_s_in%nsponge_z_l
	end do

	!##### read virtual source locations
	call from_aux("srcs","n1",n)
	if(n/=2) call seperr("expecting receiver file to have n1 = 2")

	call from_aux("srcs","n2",nsrc)
	allocate(m_s_in%srcs(2,nsrc))

	if (4 * size (m_s_in%srcs) /= sreed ("srcs", m_s_in%srcs, 4 * size (m_s_in%srcs))) &
		call seperr("model_mod: trouble reading in srcs")

	do isrc = 1, nsrc
		m_s_in%srcs(1,isrc)=m_s_in%srcs(1,isrc)+m_s_in%nsponge_x_l
		m_s_in%srcs(2,isrc)=m_s_in%srcs(2,isrc)+m_s_in%nsponge_z_l
	end do

	!validate virtual source and receiver locations
	call check_src_rec (m_s_in)


	! some global variables to save on the argument passing
	dt_glob = m_s_in%dt
	idt = 1./m_s_in%dt
	idx = 1./m_s_in%dx
	idz = 1./m_s_in%dz

end subroutine


subroutine check_src_rec (m_s_in)
	! check that all virtual sources/receivers are in the computational domain
	type(model_type)			:: m_s_in
	integer						:: irec, isrc, nrec_found, nsrc_found

	nrec_found=0
	nsrc_found=0

	! see how many receivers are in the computational domain
	nrec_found=0
	do irec = 1, size (m_s_in%recs,2)

		if (m_s_in%recs (1,irec) >= m_s_in%fx + m_s_in%nsponge_x_l .and. &
			m_s_in%recs (1,irec) <= m_s_in%fx + m_s_in%nx - m_s_in%nsponge_x_u - 1 .and. &
			m_s_in%recs (2,irec) >= m_s_in%fz + m_s_in%nsponge_z_l .and. &
			m_s_in%recs (2,irec) <= m_s_in%fz + m_s_in%nz - m_s_in%nsponge_z_u - 1) then

			nrec_found = nrec_found+1
		end if
	end do


	! see how many sources are in the computational domain
	nsrc_found=0
	do isrc = 1, size (m_s_in%srcs,2)

		if (m_s_in%srcs (1,isrc) >= m_s_in%fx + m_s_in%nsponge_x_l .and. &
			m_s_in%srcs (1,isrc) <= m_s_in%fx + m_s_in%nx - m_s_in%nsponge_x_u - 1 .and. &
			m_s_in%srcs (2,isrc) >= m_s_in%fz + m_s_in%nsponge_z_l .and. &
			m_s_in%srcs (2,isrc) <= m_s_in%fz + m_s_in%nz - m_s_in%nsponge_z_u - 1) then

			nsrc_found = nsrc_found+1
		end if
	end do

	if (nrec_found /= size (m_s_in%recs, 2) .or. nsrc_found /= size (m_s_in%srcs, 2)) then
		call seperr("Not all receivers/sources in computational domain")
	end if
end subroutine

subroutine make_wavelet (m_s_in)
	type(model_type)	:: m_s_in
	integer				:: ntw, i
	real				:: a, b, t, t2
	real, allocatable, dimension (:) :: r0, r1, r2

	ntw = 1. / m_s_in%fpeak / m_s_in%dt
	allocate (m_s_in%wavelet (ntw*2 + 1))
	allocate (r0 (ntw*2 + 1))
	allocate (r1 (ntw*2 + 1))
	allocate (r2 (ntw*2 + 1))

	r0 = 0
	r1 = 0
	r2 = 0

	b = (3.1415926 * m_s_in%fpeak)**2

	m_s_in%wavelet = 0
	do  i = 1, ntw+1
		t  = ((i-1) * m_s_in%dt)
		t2 = t*t

		! gaussian
		r0 (ntw+i)   = exp(-b * t2)
		r0 (ntw+2-i) = r0 (ntw+i)
		
		! ricker1
		r1 (ntw+i)   = -2*b*t * exp(-b * t2)
		r1 (ntw+2-i) = -r1 (ntw+i)

		! ricker2
		r2 (ntw+i)   = (1. - 2*b*t2) * exp(-b * t2)
		r2 (ntw+2-i) = r2 (ntw+i)

	end do

	r0 (1) = 0
	r1 (1) = 0
	r2 (1) = 0
	r0 (ntw*2 + 1) = 0
	r1 (ntw*2 + 1) = 0
	r2 (ntw*2 + 1) = 0

	r0 = r0 / maxval(abs(r0))
	r1 = r1 / maxval(abs(r1))
	r2 = r2 / maxval(abs(r2))

	r0 = r0 - sum(r0)/(max(1,size(r0)))
	r1 = r1 - sum(r1)/(max(1,size(r1)))
	r2 = r2 - sum(r2)/(max(1,size(r2)))

	m_s_in%wavelet = r1 ! ricker1 seems to give the least zero-frequency leftovers

	!do i = 1, ntw*2 + 1
	!	write (0,*), "waveletr0=", r0 (i)
	!	write (0,*), "waveletr1=", r1 (i)
	!	write (0,*), "waveletr2=", r2 (i)
	!end do

	deallocate (r0, r1, r2)
end subroutine

subroutine global_init (m_s_in, d_s_in, io_s_in)
	type (model_type)	:: m_s_in
	type (distrib_type)	:: d_s_in
	type (io_type)		:: io_s_in

	m_s = m_s_in
	d_s = d_s_in
	io_s = io_s_in

end subroutine

! forward and adjoint operator
function L_oper (adj, add, m, d)  result(stat)
	integer					:: stat 
	logical, intent (in)	:: adj, add 
	real, dimension (:)		:: m, d 

	call adjnull (adj, add, m, d)
	call prop_ts (adj, add, m, d)
	stat=0

end function

! data weighting
function W_oper (adj, add, m, d)  result(stat)
	integer					:: stat 
	logical, intent (in)	:: adj, add 
	real, dimension (:)		:: m, d 
	call adjnull (adj, add, m, d)
	call weight_geo_hydro (adj, add, m, d)
	stat=0
end function


subroutine prop_ts (adj, add, model, data)
	logical				:: adj, add
	real, dimension (:), target	:: model
	real, dimension (:), target	:: data
	real						:: Vsrc, Psrc, theta, vr, pi=4.*atan(1.)

	integer				:: it, dt_print, i_print

	d_s%recdata => data
	d_s%srcmodel => model

	! zero all wavefields
	d_s%all = 0.

	if (adj) then
		write (0,*), "running adjoint operator"
	else
		write (0,*), "running forward operator"
	endif

	write (0,*), ""
	! for progress bar printing
	dt_print = m_s%nt / 10
	i_print = 0
	open (unit=0, carriagecontrol='fortran')
	call progressb (0)
	do it = 1, m_s%nt
		
		!============= FORWARD OPERATOR (model --> data) =====================
		! read from source functions and generate recorded data
		!=====================================================================

		! forward modeling from point source - particle velocity source
		if (m_s%proptype .eq. source .and. it .le. size (m_s%wavelet)) then

			! the particle velocity source is supposed to be the pressure source divided by rho*Vp. Remember that rho is actually the reciprocal of rho.
			Vsrc = m_s%wavelet (it) * m_s%dt / (m_s%dx * m_s%dz) / sqrt ((d_s%lambda (m_s%ixsrc, m_s%izsrc) + 2*d_s%mu (m_s%ixsrc, m_s%izsrc)) * d_s%rho (m_s%ixsrc, m_s%izsrc))
			vr = pi / ((size (m_s%wavelet)-1) * m_s%dt)

			if (m_s%isrc_type .eq. 1) then
				d_s%vx (m_s%ixsrc, m_s%izsrc) = d_s%vx (m_s%ixsrc, m_s%izsrc) + Vsrc 

			else if (m_s%isrc_type .eq. 2) then		! monopole source in z direction
				d_s%vz (m_s%ixsrc, m_s%izsrc) = d_s%vz (m_s%ixsrc, m_s%izsrc) + Vsrc

			else if (m_s%isrc_type .eq. 3) then		! dipole source in x direction
				d_s%vx (m_s%ixsrc,   m_s%izsrc) = d_s%vx (m_s%ixsrc,   m_s%izsrc) + Vsrc 
				d_s%vx (m_s%ixsrc+1, m_s%izsrc) = d_s%vx (m_s%ixsrc+1, m_s%izsrc) - Vsrc 

			else if (m_s%isrc_type .eq. 4) then		! dipole source in z direction
				d_s%vz (m_s%ixsrc, m_s%izsrc)   = d_s%vz (m_s%ixsrc, m_s%izsrc)   + Vsrc 
				d_s%vz (m_s%ixsrc, m_s%izsrc+1) = d_s%vz (m_s%ixsrc, m_s%izsrc+1) - Vsrc 

			else if (m_s%isrc_type .eq. 5) then		! Explosive source - pure P
				d_s%vx (m_s%ixsrc,   m_s%izsrc)   = d_s%vx (m_s%ixsrc,   m_s%izsrc) - Vsrc ! top left corner
				d_s%vz (m_s%ixsrc,   m_s%izsrc)   = d_s%vz (m_s%ixsrc,   m_s%izsrc) - Vsrc
				d_s%vx (m_s%ixsrc+1, m_s%izsrc)   = d_s%vx (m_s%ixsrc+1, m_s%izsrc) + Vsrc ! bottom left corner
				d_s%vz (m_s%ixsrc+1, m_s%izsrc)   = d_s%vz (m_s%ixsrc+1, m_s%izsrc) - Vsrc
				d_s%vx (m_s%ixsrc,   m_s%izsrc+1) = d_s%vx (m_s%ixsrc,   m_s%izsrc+1) - Vsrc ! top right corner
				d_s%vz (m_s%ixsrc,   m_s%izsrc+1) = d_s%vz (m_s%ixsrc,   m_s%izsrc+1) + Vsrc
				d_s%vx (m_s%ixsrc+1, m_s%izsrc+1) = d_s%vx (m_s%ixsrc+1, m_s%izsrc+1) + Vsrc ! bottom right corner
				d_s%vz (m_s%ixsrc+1, m_s%izsrc+1) = d_s%vz (m_s%ixsrc+1, m_s%izsrc+1) + Vsrc

			else if (m_s%isrc_type .eq. 6) then		! Explosive twister source - generates isotropic P and S because this is a staggered grid implementation

				theta = vr * (it-1) * m_s%dt

				d_s%vx (m_s%ixsrc,   m_s%izsrc)   = d_s%vx (m_s%ixsrc,   m_s%izsrc)   + Vsrc * sin (theta) ! top left corner
				d_s%vz (m_s%ixsrc,   m_s%izsrc)   = d_s%vz (m_s%ixsrc,   m_s%izsrc)   - Vsrc * cos (theta)
				d_s%vx (m_s%ixsrc+1, m_s%izsrc)   = d_s%vx (m_s%ixsrc+1, m_s%izsrc)   + Vsrc * cos (theta) ! bottom left corner
				d_s%vz (m_s%ixsrc+1, m_s%izsrc)   = d_s%vz (m_s%ixsrc+1, m_s%izsrc)   + Vsrc * sin (theta)
				d_s%vx (m_s%ixsrc,   m_s%izsrc+1) = d_s%vx (m_s%ixsrc,   m_s%izsrc+1) - Vsrc * cos (theta) ! top right corner
				d_s%vz (m_s%ixsrc,   m_s%izsrc+1) = d_s%vz (m_s%ixsrc,   m_s%izsrc+1) - Vsrc * sin (theta)
				d_s%vx (m_s%ixsrc+1, m_s%izsrc+1) = d_s%vx (m_s%ixsrc+1, m_s%izsrc+1) - Vsrc * sin (theta) ! bottom right corner
				d_s%vz (m_s%ixsrc+1, m_s%izsrc+1) = d_s%vz (m_s%ixsrc+1, m_s%izsrc+1) + Vsrc * cos (theta)


			else if (m_s%isrc_type .eq. 7) then		! spindle source - funny P and S

				theta = vr * (it-1) * m_s%dt

				d_s%vx (m_s%ixsrc, m_s%izsrc) = d_s%vx (m_s%ixsrc, m_s%izsrc) + Vsrc * cos (theta) ! top left corner
				d_s%vz (m_s%ixsrc, m_s%izsrc) = d_s%vz (m_s%ixsrc, m_s%izsrc) + Vsrc * sin (theta)

			else if (m_s%isrc_type .eq. 8) then		! Implosive twister source - generates isotropic P and S because this is a staggered grid implementation

				theta = vr * (it-1) * m_s%dt

				d_s%vx (m_s%ixsrc,   m_s%izsrc)   = d_s%vx (m_s%ixsrc,   m_s%izsrc)   + Vsrc * cos (theta) ! top left corner
				d_s%vz (m_s%ixsrc,   m_s%izsrc)   = d_s%vz (m_s%ixsrc,   m_s%izsrc)   + Vsrc * sin (theta)
				d_s%vx (m_s%ixsrc+1, m_s%izsrc)   = d_s%vx (m_s%ixsrc+1, m_s%izsrc)   - Vsrc * sin (theta) ! bottom left corner
				d_s%vz (m_s%ixsrc+1, m_s%izsrc)   = d_s%vz (m_s%ixsrc+1, m_s%izsrc)   + Vsrc * cos (theta)
				d_s%vx (m_s%ixsrc,   m_s%izsrc+1) = d_s%vx (m_s%ixsrc,   m_s%izsrc+1) + Vsrc * sin (theta) ! top right corner
				d_s%vz (m_s%ixsrc,   m_s%izsrc+1) = d_s%vz (m_s%ixsrc,   m_s%izsrc+1) - Vsrc * cos (theta)
				d_s%vx (m_s%ixsrc+1, m_s%izsrc+1) = d_s%vx (m_s%ixsrc+1, m_s%izsrc+1) - Vsrc * cos (theta) ! bottom right corner
				d_s%vz (m_s%ixsrc+1, m_s%izsrc+1) = d_s%vz (m_s%ixsrc+1, m_s%izsrc+1) - Vsrc * sin (theta)

			endif

		! forward modeling from virtual source functions
		else if (m_s%proptype .eq. receiver) then
			!add geophone trace data. The trace data must be interpolated to be in modeling dt intervals, otherwise we'll reach EOF.
			call start_timer_num(tm_read)
			call read_srcs ("in_vx_src", .false., it, m_s, d_s, 1., 1, iarray1=vx_i, iarray_in=vxs_i)
			call read_srcs ("in_vz_src", .false., it, m_s, d_s, 1., 1, iarray1=vz_i, iarray_in=vzs_i)
			call stop_timer_num(tm_read)

		endif

		!advance velocities of velocity field
		call start_timer_num(tm_v4)
		call fdvsg3d_v4 (m_s%nx, m_s%nz, d_s%rho, d_s%ux, d_s%uz, d_s%vx, d_s%vz, d_s%txx, d_s%tzz, d_s%txz)
		call stop_timer_num(tm_v4)

		!do velocity boundary conditions on velocity field
		call start_timer_num(tm_sp)
		call fdvsg3d_sponge (d_s%vx, m_s%sponge_x_l, m_s%sponge_x_u, m_s%sponge_z_l, m_s%sponge_z_u)
		call fdvsg3d_sponge (d_s%vz, m_s%sponge_x_l, m_s%sponge_x_u, m_s%sponge_z_l, m_s%sponge_z_u)
		call stop_timer_num(tm_sp)

		! forward modeling from point source - pressure source
		if (m_s%proptype .eq. source .and. it .le. size (m_s%wavelet)) then
			if (m_s%isrc_type .eq. 0) then
				Psrc = m_s%Pstrength * m_s%wavelet (it) * m_s%dt / (m_s%dx * m_s%dz) ! pressure source is set to 1 Bar = 10^5 Pascal.

				d_s%txx (m_s%ixsrc, m_s%izsrc) = d_s%txx (m_s%ixsrc, m_s%izsrc) + Psrc
				d_s%tzz (m_s%ixsrc, m_s%izsrc) = d_s%tzz (m_s%ixsrc, m_s%izsrc) + Psrc

			endif
		endif

		! advance stresses of stress fields
		call start_timer_num (tm_tau4)
        call fdvsg3d_tau4 (m_s%nx, m_s%nz, d_s%lambda, d_s%mu, d_s%mu_avg, d_s%ux, d_s%uz, d_s%vx, d_s%vz, d_s%txx, d_s%tzz, d_s%txz, d_s%P, d_s%S, d_s%rot, d_s%rotc)
		call stop_timer_num (tm_tau4)

		!do stress boundary conditions on stress field
		call start_timer_num(tm_sp)
		call fdvsg3d_sponge (d_s%txx, m_s%sponge_x_l, m_s%sponge_x_u, m_s%sponge_z_l, m_s%sponge_z_u)
		call fdvsg3d_sponge (d_s%tzz, m_s%sponge_x_l, m_s%sponge_x_u, m_s%sponge_z_l, m_s%sponge_z_u)
		call fdvsg3d_sponge (d_s%txz, m_s%sponge_x_l, m_s%sponge_x_u, m_s%sponge_z_l, m_s%sponge_z_u)
		call stop_timer_num(tm_sp)

		!write out wavefields to files
		call start_timer_num (tm_write)
		call write_out_files (it, m_s, d_s, io_s, adj)
		call stop_timer_num  (tm_write)

		! progress
		if ( mod (it, dt_print) == 0) then
			i_print = i_print + 1
			call progressb (i_print)
		endif

	end do
	write (0,*), ""

	return
end subroutine

! =============================================================
!     update vx, vz using 4th order staggered grid
! =============================================================

subroutine fdvsg3d_v4 (nx, nz, rho, ux, uz, vx, vz, txx, tzz, txz)
	real, parameter			:: c0=9./8., c1=1./24.
	real					:: cx0, cx1, cz0, cz1
	integer					:: ix, iz, nx, nz
	real, dimension (:,:)	:: rho, vx, ux, uz, vz, txx, tzz, txz
	real					:: bx, bz, dx_txx, dz_tzz, dz_txz, dx_txz
	
	cx0=idx*c0
	cx1=idx*c1
	cz0=idz*c0
	cz1=idz*c1

!$OMP PARALLEL private (iz, ix, bx, bz, dx_txx, dz_txz, dx_txz, dz_tzz)
!$OMP DO
	do iz = 3, nz-2
		do ix = 3, nx-2

			bx    	= 0.5 * (rho (ix, iz) + rho (ix+1, iz))
			dx_txx	= cx0 * (txx (ix+1, iz) - txx (ix, iz  )) - cx1 * (txx (ix+2, iz)   - txx (ix-1, iz  ))
			dz_txz	= cz0 * (txz (ix,   iz) - txz (ix, iz-1)) - cz1 * (txz (ix,   iz+1) - txz (ix,   iz-2))

			vx (ix, iz) = vx (ix, iz) + dt_glob * bx * (dx_txx + dz_txz)
			ux (ix, iz) = ux (ix, iz) + dt_glob * vx (ix, iz)

			bz    	= 0.5 * (rho (ix, iz) + rho (ix, iz+1))
			dx_txz	= cx0 * (txz (ix, iz  ) - txz (ix-1, iz)) - cx1 * (txz (ix+1, iz)   - txz (ix-2, iz  ))
			dz_tzz	= cz0 * (tzz (ix, iz+1) - tzz (ix,   iz)) - cz1 * (tzz (ix,   iz+2) - tzz (ix,   iz-1))
 
			vz (ix, iz) = vz (ix, iz) + dt_glob * bz * (dz_tzz + dx_txz)
			uz (ix, iz) = uz (ix, iz) + dt_glob * vz (ix, iz)

		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

! ===============================================================
!     tau4
!     update txx, tzz, txz using 4th order
! ===============================================================
subroutine fdvsg3d_tau4 (nx, nz, lambda, mu, mu_avg, ux, uz, vx, vz, txx, tzz, txz, P, S, rot, rotc)
	real,  parameter		:: c0 = 9./8., c1=1./24.
	real					:: cx0, cx1, cz0, cz1
	integer					:: ix, iz, nz, nx
	real					:: r2u, rta
	real					:: dx_vx, dz_vx, dx_vz, dz_vz, dx_uz, dz_ux
	real, dimension (:,:)	:: lambda, mu, mu_avg, ux, uz, vx, vz, txx, tzz, txz, P, S, rot, rotc

	cx0=idx*c0
	cx1=idx*c1
	cz0=idz*c0
	cz1=idz*c1

	! calculate normal stresses txx, tzz
!$OMP PARALLEL private (ix, iz, r2u, rta, dx_vx, dz_vz, dz_vx, dx_vz)
!$OMP DO
	do iz = 3, nz-2
		do ix = 3, nx-2
			r2u = lambda (ix, iz) + 2 * mu (ix, iz)
			rta = lambda (ix, iz)

			dx_vx = cx0 * (vx (ix, iz) - vx (ix-1, iz  )) - cx1 * (vx (ix+1, iz  ) - vx (ix-2, iz  ))
			dz_vz = cz0 * (vz (ix, iz) - vz (ix,   iz-1)) - cz1 * (vz (ix,   iz+1) - vz (ix,   iz-2))
				
			txx (ix, iz) = txx (ix, iz) + dt_glob * (r2u * dx_vx + rta * dz_vz)
			tzz (ix, iz) = tzz (ix, iz) + dt_glob * (r2u * dz_vz + rta * dx_vx)
			!P   (ix, iz) = 0.5 * ((r2u + rta) * (dx_vx + dz_vz)) ! P wave amplitude with dimensions ? trace of stress matrix?
			P   (ix, iz) = P (ix, iz) + 0.5 * (txx (ix, iz) + tzz (ix, iz)) ! integrated P wave amplitude with dimensions ? trace of stress matrix?
			!P   (ix, iz) = P (ix, iz) + dt_glob * (dx_vx + dz_vz) ! divergence - P wave amplitude without dimensions

			dz_vx = cz0 * (vx (ix,   iz+1) - vx (ix, iz)) - cz1 * (vx (ix,   iz+2) - vx (ix, iz-1))
			dx_vz = cx0 * (vz (ix+1, iz  ) - vz (ix, iz)) - cx1 * (vz (ix+2, iz  ) - vz (ix-1, iz))

			txz (ix, iz) = txz (ix, iz) + dt_glob * mu_avg (ix, iz) * (dx_vz + dz_vx)

			!S (ix, iz) = mu_avg (ix, iz) * (dz_vx - dx_vz) ! S wave amplitude with dimensions?
			S (ix, iz) = S (ix, iz) + dt_glob * mu_avg (ix, iz) * (dz_vx - dx_vz) ! integrated S wave amplitude with dimensions?
			!S (ix, iz) = S (ix, iz) + dt_glob * (dz_vx - dx_vz) ! curl - S wave amplitude without dimensions

			dz_ux = cz0 * (ux (ix,   iz+1) - ux (ix, iz)) - cz1 * (ux (ix,   iz+2) - ux (ix, iz-1))
			dx_uz = cx0 * (uz (ix+1, iz  ) - uz (ix, iz)) - cx1 * (uz (ix+2, iz  ) - uz (ix-1, iz))
			rot (ix, iz)   = idt * ((dz_ux - dx_uz) - rotc (ix, iz)) ! first time derivative of rotation
			rotc (ix, iz) = dz_ux - dx_uz ! pure curl - rotation?

		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

! Weight the data residual, so as not to include the hydrophone data in the inversion (for the moment)
subroutine weight_geo_hydro (adj, add, model, data)
	logical				:: adj, add
	real, dimension (:), target	:: model
	real, dimension (:), target	:: data
	integer						:: nrecs, nsrcs

	nrecs = size (m_s%recs,2)
	nsrcs = size (m_s%srcs,2)

	! make hydrophone data = 0, so that it doesn't get into the model gradient calculation
	if (adj) then
		model (1 + nrecs * m_s%nt * vxr_i : nrecs * m_s%nt * (vxr_i + 1)) = 1. *  data (1 + nrecs * m_s%nt * vxr_i : nrecs * m_s%nt * (vxr_i + 1))
		model (1 + nrecs * m_s%nt * vzr_i : nrecs * m_s%nt * (vzr_i + 1)) = 1. *  data (1 + nrecs * m_s%nt * vzr_i : nrecs * m_s%nt * (vzr_i + 1))
		model (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1)) = 0. *  data (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1))
	else
	! forward: both model and data are of data dimensions. I want to not use the hydrophone at the moment, so it is given a weight of 0.
		data (1 + nrecs * m_s%nt * vxr_i : nrecs * m_s%nt * (vxr_i + 1)) = 1. *  model (1 + nrecs * m_s%nt * vxr_i : nrecs * m_s%nt * (vxr_i + 1))
		data (1 + nrecs * m_s%nt * vzr_i : nrecs * m_s%nt * (vzr_i + 1)) = 1. *  model (1 + nrecs * m_s%nt * vzr_i : nrecs * m_s%nt * (vzr_i + 1))
		data (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1)) = 0. *  model (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1))

	endif

end subroutine

! ======================================================================
!     dump velocity near boundaries
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================
subroutine fdvsg3d_sponge (array, xsponge_l, xsponge_u, zsponge_l, zsponge_u)
	real :: array(:,:)
	real :: xsponge_l (:), xsponge_u (:), zsponge_l (:), zsponge_u (:)
	integer :: ix,iz


	!==== absorb in X direction
!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz = 1, m_s%nz
		do ix = 1, m_s%nsponge_x_l
			array (ix,iz) = array (ix, iz) * xsponge_l (ix)
		end do
		do ix = 1, m_s%nsponge_x_u
			array (m_s%nx+1-ix, iz) = array (m_s%nx+1-ix, iz) * xsponge_u (ix)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	!==== absorb in Z direction
!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz = 1, m_s%nsponge_z_l
		do ix = 1 + m_s%nsponge_x_l, m_s%nx - m_s%nsponge_x_u
			array(ix,iz) = array (ix, iz) * zsponge_l (iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	
!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,m_s%nsponge_z_u
		do ix = 1 + m_s%nsponge_x_l, m_s%nx - m_s%nsponge_x_u
			array (ix, m_s%nz+1-iz) = array (ix, m_s%nz+1-iz) * zsponge_u (iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

subroutine mk_sponge (m_s_in)
	type (model_type)	:: m_s_in
	real				:: a0, damper
	integer				:: i, nx_l, nx_u, nz_l, nz_u
	real				:: pi = 4. * atan(1.)

	call from_param ("surf_type", m_s_in%surf_type, 0)
	call from_param ("xsponge_l", nx_l, 10)
	call from_param ("xsponge_u", nx_u, 10)
	call from_param ("zsponge_l", nz_l, 10)
	call from_param ("zsponge_u", nz_u, 10)
	call from_param ("damper",  damper, 0.6)


	! lower x boundary
	if (m_s_in%surf_type .eq. free) then	! free surface: add just one cell from above, with no damping.
		m_s_in%nsponge_x_l = 1
		allocate (m_s_in%sponge_x_l (1))
		m_s_in%sponge_x_l (1) = 1.

	else							! rigid surface: do absorbtion as usual.
		m_s_in%nsponge_x_l = nx_l
		if (nx_l .gt. 0.) then
			a0 = damper / real (nx_l)
	
			allocate (m_s_in%sponge_x_l (nx_l))
			do i = 1, nx_l
				m_s_in%sponge_x_l (i) = exp(- (a0 * real (nx_l + 1 - i))**2)
			end do
		endif
	endif

	! upper x boundary
	m_s_in%nsponge_x_u = nx_u
	if (nx_u .gt. 0.) then
		a0 = damper / real (nx_u)

		allocate (m_s_in%sponge_x_u (nx_u))
		do i = 1, nx_u
			m_s_in%sponge_x_u (i) = exp(- (a0 * real (nx_u + 1 - i))**2)
		end do
	endif

	! lower z boundary
	m_s_in%nsponge_z_l = nz_l
	if (nz_l .gt. 0.) then
		a0 = damper / real (nz_l)

		allocate (m_s_in%sponge_z_l (nz_l))
		do i = 1, nz_l
			m_s_in%sponge_z_l (i) = exp (- (a0 * real (nz_l + 1 - i))**2)
		end do
	endif

	! upper z boundary
	m_s_in%nsponge_z_u = nz_u
	if (nz_u .gt. 0.) then
		a0 = damper / real (nz_u)

		allocate (m_s_in%sponge_z_u (nz_u))
		do i = 1, nz_u
			m_s_in%sponge_z_u (i) = exp (- (a0 * real (nz_u + 1 - i))**2)
		end do
	endif

end subroutine

subroutine progressb (j)
	integer	:: j, k
	character (len=12) :: bar="|          |"

	if (j <= 10) then
		do k = 1, j
			bar (1+k:1+k) = "*"
		enddo
		do k = j+1, 10
			bar (1+k:1+k) = " "
		enddo
		write (unit=0, fmt="(a1,a1,a12)") '+', char(13), bar
	endif

	return
end subroutine

integer function pad2 (n)
	integer	:: n
	pad2 = 1

	do while (pad2 .lt. n)
		pad2 = pad2*2
	end do

	return
end function pad2


end module
